indexing

	description: "[
		Create a separate thread wherein the service runs. As a thread has its
		own message queue, it is very easy for the controller function to
		communicate with the service. The controller function runs in the thread
		that called StartServiceCtrlHandler (usually the primary thread). Now
		the controller function can just send a message to this threads message
		queue.
]"

	author: "Berend de Boer"
	copyright:   "Copyright (c) 2003, Berend de Boer"
	license:     "Eiffel Forum Freeware License v2 (see forum.txt)"


class

	NT_SERVICE_THREAD


inherit

	THREAD
		rename
			thread_id as thread_handle
		export
			{ANY} terminated
		end

	NT_SERVICE_DEBUGGING
		export
			{NONE} all
		end

	WINSVC_CONSTANTS
		export
			{NONE} all
		end

	EXCEPTIONS
		export
			{NONE} all
		end


creation

	make


feature {NONE} -- Initialization

	make (a_service: like service) is
		require
			service_not_void: a_service /= Void
		do
			service := a_service
			create msg.make
		end


feature -- Access

	thread_id: INTEGER
			-- win32 thread identifier.

	service: NT_SERVICE
			-- The service for which this thread is executing.


feature -- Execution

	execute is
		local
			retried: BOOLEAN
		do
			if not retried then
				-- Create a message queue so we can receive the requests send
				-- by SERVICE.controller.
				thread_id := cwin_get_current_thread_id
				msg.create_thread_message_queue
				debug ("service")
					log ("NT_SERVICE_THREAD.execute: start service")
				end
				service.start
				if service.current_state = service_running then
					debug ("service")
						log ("NT_SERVICE_THREAD.execute: execute service")
					end
					service.execute
				else
					service.stop
				end
			end
			terminated := True
		ensure then
			terminated: terminated
		rescue
			debug ("service")
				log ("NT_SERVICE_THREAD.execute: rescue")
				if exception_trace /= Void then
					log (exception_trace)
				end
			end
			retried := True
			retry
		end

	process_requests (wait_for_message: BOOLEAN) is
		local
			retried: BOOLEAN
			action: INTEGER
		do
			if not retried then
				debug ("service")
					--log ("NT_SERVICE_THREAD.process_requests: begin")
				end
				if not terminated or else wait_for_message then
					from
						if wait_for_message then
							msg.get_all
						else
							msg.peek_all
						end
					until
						terminated or else
						not msg.last_boolean_result
					loop
						if msg.is_thread_message then
							if msg.message = service.cm_service_control_code then
								-- protect against exceptions here
								action := msg.wparam
								dispatch_service_control_code (action)
							else
								msg.dispatch
							end
						else
							msg.dispatch
						end
						if not terminated and then wait_for_message then
							msg.get_all
						end
					end
				end
				debug ("service")
					--log ("NT_SERVICE_THREAD.process_requests: end")
				end
			end
		rescue
			debug ("service")
				log ("NT_SERVICE_THREAD.process_requests: rescue, msg: " + msg.message.out)
				if exception_trace /= Void then
					log (exception_trace)
				end
			end
			retried := True
			retry
		end

	terminate is
		do
			terminated := True
		end


feature {NONE} -- Implementation

	dispatch_service_control_code (action: INTEGER) is
			-- Call proper method in `service', based on `action'.
		local
			retried: BOOLEAN
		do
			if not retried then
				debug ("service")
					log ("NT_SERVICE_THREAD.dispatch_service_control_code: begin")
				end
				inspect action
				when service_control_stop then
					if
						service.current_state /= service_stop_pending and then
						service.current_state /= service_stopped
				then
					service.stop
					end
				when service_control_shutdown then
					service.shutdown
				when service_control_interrogate then
					service.interrogate
				else
					--ignore other actions
				end
			end
		rescue
			retried := True
			retry
		end

	msg: X_WEL_MSG
			-- Access to thread message queue.


invariant

	msg_not_void: msg /= Void
	service_not_void: service /= Void

end
